import datetime
import dateutil.relativedelta

def get_current_time():
    # TODO return timezone or utc? get config from user?
    return datetime.datetime.utcnow()

def get_time_delta(**kwargs):
    return dateutil.relativedelta.relativedelta(**kwargs)




import re
import yaml
from pyworker.util import get_current_time, get_time_delta

_job_class_registry = {}

def _register_class(target_class):
    _job_class_registry[target_class.__name__] = target_class


class Meta(type):
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        _register_class(cls)
        return cls


class Job(object):
    """docstring for Job"""
    __metaclass__ = Meta
    def __init__(self, class_name, database, logger,
        job_id, attempts=0, max_attempts=1, attributes=None):
        super(Job, self).__init__()
        self.class_name = class_name
        self.database = database
        self.logger = logger
        self.job_id = job_id
        self.attempts = attempts
        self.max_attempts = max_attempts
        self.attributes = attributes

    def __str__(self):
        return "%s: %s" % (self.__class__.__name__, str(self.__dict__))

    @classmethod
    def from_row(cls, job_row, max_attempts, database, logger):
        '''job_row is a tuple of (id, attempts, handler)'''
        def extract_class_name(line):
            # TODO cache regex
            regex = re.compile('object: !ruby/object:(.+)')
            match = regex.match(line)
            if match:
                return match.group(1)
            else:
                return None

        def extract_attributes(lines):
            attributes = []
            collect = False
            for line in lines:
                if line.startswith('  raw_attributes:'):
                    collect = True
                elif not line.startswith('    '):
                    if collect:
                        break
                elif collect:
                    attributes.append(line)
            return attributes

        job_id, attempts, handler = job_row
        handler = handler.splitlines()

        class_name = extract_class_name(handler[1])
        logger.debug("Found Job %d with class name: %s" % (job_id, class_name))
        try:
            target_class = _job_class_registry[class_name]
        except KeyError:
            return Job(class_name=class_name, logger=logger,
                max_attempts=max_attempts,
                job_id=job_id, attempts=attempts, database=database)

        attributes = extract_attributes(handler[2:])
        logger.debug("Found attributes: %s" % str(attributes))

        stripped = '\n'.join(['object:', '  attributes:'] + attributes)
        payload = yaml.load(stripped)
        logger.debug("payload object: %s" % str(payload))

        return target_class(class_name=class_name, logger=logger,
            job_id=job_id, attempts=attempts, database=database,
            max_attempts=max_attempts,
            attributes=payload['object']['attributes'])

    def before(self):
        self.logger.debug("Running Job.before hook")

    def after(self):
        self.logger.debug("Running Job.after hook")

    def error(self, error):
        self.logger.debug("Running Job.error hook")

    def failure(self, error):
        self.logger.debug("Running Job.failure hook")

    def success(self):
        self.logger.debug("Running Job.success hook")

    def set_error_unlock(self, error):
        self.logger.error('Job %d raised error: %s' % (self.job_id, error))
        # run error hook
        self.error(error)
        self.attempts += 1
        now = get_current_time()
        setters = [
            'locked_at = null',
            'locked_by = null',
            'attempts = %d' % self.attempts,
            'last_error = %s'
        ]
        values = [
            error
        ]
        if self.attempts >= self.max_attempts:
            # set failed_at = now
            setters.append('failed_at = %s')
            values.append(now)
            self.failure(error)
        else:
            # set new exponential run_at
            setters.append('run_at = %s')
            delta = (self.attempts**4) + 5
            values.append(str(now + get_time_delta(seconds=delta)))
        query = 'UPDATE delayed_jobs SET %s WHERE id = %d' % \
            (', '.join(setters), self.job_id)
        self.logger.debug('set error query: %s' % query)
        self.logger.debug('set error values: %s' % str(values))
        self.database.cursor().execute(query, tuple(values))
        self.database.commit()

    def remove(self):
        self.logger.debug('Job %d finished successfully' % self.job_id)
        query = 'DELETE FROM delayed_jobs WHERE id = %d' % self.job_id
        self.database.cursor().execute(query)
        self.database.commit()



import os, sys, signal, traceback
import time
from contextlib import contextmanager
from pyworker.db import DBConnector
from pyworker.job import Job
from pyworker.logger import Logger
from pyworker.util import get_current_time, get_time_delta

class TimeoutException(Exception): pass
class TerminatedException(Exception): pass

class Worker(object):
    def __init__(self, dbstring, logger=None):
        super(Worker, self).__init__()
        self.logger = Logger(logger)
        self.logger.info('Starting pyworker...')
        self.database = DBConnector(dbstring, self.logger)
        self.sleep_delay = 10
        self.max_attempts = 3
        self.max_run_time = 3600
        self.queue_names = 'default'
        hostname = os.uname()[1]
        pid = os.getpid()
        self.name = 'host:%s pid:%d' % (hostname, pid)

    @contextmanager
    def _time_limit(self, seconds):
        def signal_handler(signum, frame):
            raise TimeoutException(('Execution expired. Either do ' + \
                'the job faster or raise max_run_time > %d seconds') % \
                self.max_run_time)
        signal.signal(signal.SIGALRM, signal_handler)
        signal.alarm(seconds)
        try:
            yield
        finally:
            signal.alarm(0)

    @contextmanager
    def _terminatable(self):
        def signal_handler(signum, frame):
            signal_name = 'SIGTERM' if signum == 15 else 'SIGINT'
            self.logger.info('Received signal: %s' % signal_name)
            raise TerminatedException(signal_name)
        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        yield

    def run(self):
        # continuously check for new jobs on specified queue from db
        self._cursor = self.database.connect().cursor()
        with self._terminatable():
            while True:
                self.logger.debug('Picking up jobs...')
                job = self.get_job()
                self._current_job = job # used in signal handlers
                start_time = time.time()
                try:
                    if type(job) == Job:
                        raise ValueError(('Unsupported Job: %s, please import it ' \
                            + 'before you can handle it') % job.class_name)
                    elif job is not None:
                        self.logger.info('Running Job %d' % job.job_id)
                        with self._time_limit(self.max_run_time):
                            job.before()
                            job.run()
                            job.after()
                        job.success()
                        job.remove()
                    time.sleep(self.sleep_delay)
                except Exception as exception:
                    if job is not None:
                        error_str = traceback.format_exc()
                        job.set_error_unlock(error_str)
                    if type(exception) == TerminatedException:
                        break
                finally:
                    if job is not None:
                        time_diff = time.time() - start_time
                        self.logger.info('Job %d finished in %d seconds' % \
                            (job.job_id, time_diff))
            
            self.database.disconnect()

    def get_job(self):
        def get_job_row():
            now = get_current_time()
            expired = now - get_time_delta(seconds=self.max_run_time)
            now, expired = str(now), str(expired)
            queues = self.queue_names.split(',')
            queues = ', '.join(["'%s'" % q for q in queues])
            query = '''
            UPDATE delayed_jobs SET locked_at = '%s', locked_by = '%s'
            WHERE id IN (SELECT delayed_jobs.id FROM delayed_jobs
                WHERE ((run_at <= '%s'
                AND (locked_at IS NULL OR locked_at < '%s')
                OR locked_by = '%s') AND failed_at IS NULL)
                AND delayed_jobs.queue IN (%s)
            ORDER BY priority ASC, run_at ASC LIMIT 1 FOR UPDATE) RETURNING
                id, attempts, handler
            ''' % (now, self.name, now, expired, self.name, queues)
            self.logger.debug('query: %s' % query)
            self._cursor.execute(query)
            return self._cursor.fetchone()

        job_row = get_job_row()
        if job_row:
            return Job.from_row(job_row, max_attempts=self.max_attempts,
                database=self.database, logger=self.logger)
        else:
            return None



import sys
major_version = sys.version_info.major
if major_version == 2:
    from urlparse import urlparse, parse_qs
elif major_version == 3:
    from urllib.parse import urlparse, parse_qs
import psycopg2

class DBConnector(object):
    def __init__(self, dbstring, logger):
        super(DBConnector, self).__init__()
        url = urlparse(dbstring)
        self._database = url.path[1:]
        self._username = url.username.replace('%40', '@')
        self._passwd = url.password
        self._host = url.hostname
        self._port = url.port
        qs = parse_qs(url.query)
        self._sslmode = qs.get('sslmode', ['prefer'])[0]
        self.logger = logger

    def connect(self): 
        self._connection = psycopg2.connect(database=self._database,
            user=self._username, password=self._passwd,
            host=self._host, port=self._port, sslmode=self._sslmode)
        cursor = self._connection.cursor()
        self.logger.info("Connected to database")
        return self
        
    def cursor(self):
        return self._connection.cursor()

    def commit(self):
        self._connection.commit()

    def disconnect(self):
        self._connection.close(); 
        self.logger.info("Disconnected from database")
